/*
    This file is part of the KDE libraries
    SPDX-FileCopyrightText: 2020 Aleksei Nikiforov <darktemplar@basealt.ru>

    SPDX-License-Identifier: LGPL-2.0-or-later
*/

#include <QCoreApplication>
#include <QTextStream>
#include <QString>
#include <QMap>
#include <QLatin1Char>

#include <KConfigSerialization>
#include <KConfig>
#include <KConfigGroup>

enum class event_type
{
    write_entry = 0,
    write_path_entry,
    write_path_array_entry,
    delete_entry,
    delete_group,
};

static const QMap<QString, event_type> string_to_event_type_map =
{
    { QStringLiteral("write_entry"),            event_type::write_entry },
    { QStringLiteral("write_path_entry"),       event_type::write_path_entry },
    { QStringLiteral("write_path_array_entry"), event_type::write_path_array_entry },
    { QStringLiteral("delete_entry"),           event_type::delete_entry },
    { QStringLiteral("delete_group"),           event_type::delete_group },
};

int main(int argc, char **argv)
{
    QCoreApplication app(argc, argv);
    app.setApplicationVersion(QStringLiteral("1.0"));

    QTextStream stdin_stream(stdin, QIODevice::ReadOnly);

    bool in_message = false;
    QMap<QString, QString> message_data;

    while (!stdin_stream.atEnd())
    {
        QString line = stdin_stream.readLine();

        if (in_message)
        {
            if (line == QStringLiteral("message_end"))
            {
                do
                {
                    if (message_data[QStringLiteral("version")] != QStringLiteral("1"))
                    {
                        break;
                    }

                    event_type event_type_value;
                    QString filename;
                    QStandardPaths::StandardLocation standard_location;
                    KConfigBase::WriteConfigFlags write_config_flags;
                    QStringList group_names;
                    QString entry_name;
                    QString entry_value;

                    // TODO: process uid/username, use it for optional filtering

                    {
                        auto iter = string_to_event_type_map.find(message_data[QStringLiteral("event")]);
                        if (iter == string_to_event_type_map.end())
                        {
                            break;
                        }

                        event_type_value = iter.value();
                    }

                    {
                        auto iter = message_data.find(QStringLiteral("filename"));
                        if (iter == message_data.end())
                        {
                            break;
                        }

                        filename = iter.value();
                    }

                    {
                        auto iter = message_data.find(QStringLiteral("location_type"));;
                        if (iter == message_data.end())
                        {
                            break;
                        }

                        bool ok = false;
                        standard_location = deserializeStandardLocation(iter.value(), ok);
                        if (!ok)
                        {
                            break;
                        }
                    }

                    {
                        auto iter = message_data.find(QStringLiteral("flags"));;
                        if (iter == message_data.end())
                        {
                            break;
                        }

                        bool ok = false;
                        write_config_flags = deserializeWriteConfigFlags(iter.value(), ok);
                        if (!ok)
                        {
                            break;
                        }
                    }

                    {
                        auto iter = message_data.find(QStringLiteral("group_name"));;
                        if (iter == message_data.end())
                        {
                            break;
                        }

                        group_names = iter.value().split(QLatin1Char('|'));
                        if (group_names.isEmpty())
                        {
                            break;
                        }
                    }

                    {
                        auto iter = message_data.find(QStringLiteral("entry_name"));;
                        if ((iter == message_data.end())
                            && ((event_type_value == event_type::write_entry)
                                || (event_type_value == event_type::write_path_entry)
                                || (event_type_value == event_type::write_path_array_entry)
                                || (event_type_value == event_type::delete_entry)))
                        {
                            break;
                        }

                        if (iter != message_data.end())
                        {
                            entry_name = iter.value();
                        }
                    }

                    {
                        auto iter = message_data.find(QStringLiteral("entry_value"));;
                        if ((iter == message_data.end())
                            && ((event_type_value == event_type::write_entry)
                                || (event_type_value == event_type::write_path_entry)))
                        {
                            break;
                        }

                        if (iter != message_data.end())
                        {
                            entry_value = iter.value();
                        }
                    }

                    KConfig config(filename, KConfig::FullConfig, standard_location);
                    config.setUpdateNotificationEnabled(false);

                    auto group_name_iter = group_names.begin();
                    KConfigGroup config_group = config.group(*group_name_iter);

                    for (++group_name_iter; group_name_iter != group_names.end(); ++group_name_iter)
                    {
                        config_group = config_group.group(*group_name_iter);
                    }

                    switch (event_type_value)
                    {
                    case event_type::write_entry:
                        config_group.writeEntry(entry_name, entry_value, write_config_flags);
                        break;

                    case event_type::write_path_entry:
                        config_group.writePathEntry(entry_name, entry_value, write_config_flags);
                        break;

                    case event_type::write_path_array_entry:
                        {
                            QStringList entries_array;

                            for (int idx = 1; ; ++idx)
                            {
                                auto iter = message_data.find(QStringLiteral("entry_path_%1").arg(idx));
                                if (iter == message_data.end())
                                {
                                    break;
                                }

                                entries_array.push_back(iter.value());
                            }

                            config_group.writePathEntry(entry_name, entries_array, write_config_flags);
                        }
                        break;

                    case event_type::delete_entry:
                        config_group.deleteEntry(entry_name, write_config_flags);
                        break;

                    case event_type::delete_group:
                        config_group.deleteGroup(write_config_flags);
                        break;
                    }
                } while (false);

                in_message = false;
                message_data.clear();
            }
            else
            {
                // if '=' character is present and not first character in line
                int index = line.indexOf(QLatin1Char('='));
                if (index > 0)
                {
                    message_data[line.left(index)] = line.mid(index + 1);
                }
            }
        }
        else
        {
            if (line == QStringLiteral("message_begin"))
            {
                in_message = true;
            }
        }
    }

    return 0;
}
