/*
    This file is part of the KDE libraries
    SPDX-FileCopyrightText: 2020 Aleksei Nikiforov <darktemplar@basealt.ru>

    SPDX-License-Identifier: LGPL-2.0-or-later
*/

#include <stdio.h>
#include <unistd.h>
#include <sys/socket.h>
#include <inttypes.h>
#include <signal.h>

#include <QCoreApplication>
#include <QSocketNotifier>
#include <QDBusConnection>
#include <QDBusMessage>
#include <QStringList>
#include <QString>

#include "kconf_watcher_p.h"

static int sigintFd[2];
static int sigtermFd[2];

void intSignalHandler(int)
{
    char a = 1;
    ::write(sigintFd[0], &a, sizeof(a));
}

void termSignalHandler(int)
{
    char a = 1;
    ::write(sigtermFd[0], &a, sizeof(a));
}

void qtHandleSigInt(QSocketNotifier &notifier)
{
    notifier.setEnabled(false);

    char tmp;
    ::read(sigintFd[1], &tmp, sizeof(tmp));

    QCoreApplication::quit();

    notifier.setEnabled(true);
}

void qtHandleSigTerm(QSocketNotifier &notifier)
{
    notifier.setEnabled(false);

    char tmp;
    ::read(sigtermFd[1], &tmp, sizeof(tmp));

    QCoreApplication::quit();

    notifier.setEnabled(true);
}

void DBusPrinter::messageSlot(const QStringList &values)
{
    static qint64 message_idx = 1;
    fprintf(stdout, "message %" PRIi64 "\n", message_idx++);
    fprintf(stdout, "message_begin\n");

    for (const auto &item: values)
    {
        fprintf(stdout, "%s\n", item.toLocal8Bit().data());
    }

    fprintf(stdout, "message_end\n\n");
    fflush(stdout);
}

int main(int argc, char **argv)
{
    QCoreApplication app(argc, argv);
    app.setApplicationVersion(QStringLiteral("1.0"));

    if (!QDBusConnection::sessionBus().isConnected())
    {
        fprintf(stderr,
            "Cannot connect to the D-Bus session bus.\n"
            "Please check your system settings and try again.\n");

        return -1;
    }

    if (::socketpair(AF_UNIX, SOCK_STREAM, 0, sigintFd))
    {
        fprintf(stderr, "Couldn't create INT socketpair\n");
        return -1;
    }

    if (::socketpair(AF_UNIX, SOCK_STREAM, 0, sigtermFd))
    {
        fprintf(stderr, "Couldn't create TERM socketpair\n");
        return -1;
    }

    QSocketNotifier sigIntNotifier(sigintFd[1], QSocketNotifier::Read);
    QObject::connect(&sigIntNotifier, &QSocketNotifier::activated, [&sigIntNotifier] () { qtHandleSigInt(sigIntNotifier); });

    QSocketNotifier sigTermNotifier(sigtermFd[1], QSocketNotifier::Read);
    QObject::connect(&sigTermNotifier, &QSocketNotifier::activated, [&sigTermNotifier] () { qtHandleSigTerm(sigTermNotifier); });

    {
        struct sigaction sigaction_int, sigaction_term;

        sigaction_int.sa_handler = &intSignalHandler;
        sigemptyset(&sigaction_int.sa_mask);
        sigaction_int.sa_flags = 0;
        sigaction_int.sa_flags |= SA_RESTART;
        sigaction(SIGINT, &sigaction_int, 0);

        sigaction_term.sa_handler = &termSignalHandler;
        sigemptyset(&sigaction_term.sa_mask);
        sigaction_term.sa_flags = 0;
        sigaction_term.sa_flags |= SA_RESTART;
        sigaction(SIGTERM, &sigaction_term, 0);
    }

    DBusPrinter dbus_printer;
    QDBusConnection::sessionBus().connect(
        QString(),
        QStringLiteral("/"),
        QStringLiteral("org.alt.kde.kconfig.extra.notify"),
        QStringLiteral("ConfigChanged"),
        &dbus_printer,
        SLOT(messageSlot(QStringList)));

    return app.exec();
}
